#!/usr/local/bin/ruby
# coding: utf-8

# ローターとプラグボードの配線(Enigmaクラスの中で定義した方がいいけど)
#
# https://www.cryptomuseum.com/crypto/enigma/wiring.htm
# 
# Enigma I German Army and Air Force (Wehrmacht, Luftwaffe)
# Enigma M3 German Navy (Kriegsmarine)
# Rotor # abcdefghijklmnopqrstuvwxyz      Date Introduced Model Name & Number
# ETW     ABCDEFGHIJKLMNOPQRSTUVWXYZ      (plug board)
# I       EKMFLGDQVZNTOWYHXUSPAIBRCJ      1930    Enigma I
# II      AJDKSIRUXBLHWTMCQGZNPYFVOE      1930    Enigma I
# III     BDFHJLCPRTXVZNYEIWGAKMUSQO      1930    Enigma I
# IV      ESOVPZJAYQUIRHXLNFTGKDCMWB      December 1938   M3 Army
# V       VZBRGITYUPSDNHLXAWMJQOFECK      December 1938   M3 Army
# VI      JPGVOUMFYQBENHZRDKASXLICTW      1939    M3 & M4 Naval (FEB 1942)
# VII     NZJHGRCXMYSWBOUFAIVLPEKQDT      1939    M3 & M4 Naval (FEB 1942)
# VIII    FKQHTLXOCBJSPDZRAMEWNIUYGV      1939    M3 & M4 Naval (FEB 1942)
# UKW-A   EJMZALYXVBWFCRQUONTSPIKHGD      (Enigma I) (reflector)
# UKW-B   YRUHQSLDPXNGOKMIEBFZCWVJAT	  (Enigma M3) (reflector)
# UKW-C   FVPJIAOYEDRZXWGCTKUQSBNMHL      (Enigma M3) (reflector)
#
# 
# ローターIは初期位置aで入力aをEに、bをKに、cをMに変換する
# ローターIが1段階回ると、入力aをKに変換する
# ローターIは位置qからrに回るときにだけ次のローターを回す
#
# 暗号化・復号化のキーに付いて
# 1. どのローターをどの順にセットするか
# 1a. ローターのリング設定(入力オフセット: 0のときa→E、1のときa→K、...25)
#     ノッチの位置は変わらない
# 2. どのリフレクタを使うか
# 3. プラグボードの設定  (以上が暗号書で決まっている)
# 4. ローターの開始位置  (通信員が適当に決める)
# 暗号書で決まっている部分は、月、日、時刻、部隊などから指定

# ローターの配線
#              abcdefghijklmnopqrstuvwxyz
ROTOR1_WIRE = "EKMFLGDQVZNTOWYHXUSPAIBRCJ" # リング設定0のときaをEに変換
ROTOR2_WIRE = "AJDKSIRUXBLHWTMCQGZNPYFVOE"
ROTOR3_WIRE = "BDFHJLCPRTXVZNYEIWGAKMUSQO"
ROTOR4_WIRE = "ESOVPZJAYQUIRHXLNFTGKDCMWB"
ROTOR5_WIRE = "VZBRGITYUPSDNHLXAWMJQOFECK"
ROTOR6_WIRE = "JPGVOUMFYQBENHZRDKASXLICTW"
ROTOR7_WIRE = "NZJHGRCXMYSWBOUFAIVLPEKQDT"
ROTOR8_WIRE = "FKQHTLXOCBJSPDZRAMEWNIUYGV"

# ノッチの位置
ROTOR1_NOTCH = "q" # リング設定0のとき入力q→rに移行するとき次のローターを回す
ROTOR2_NOTCH = "e" # ノッチの固定ネジが円周上に4本あるので4種類選べそうだが??
ROTOR3_NOTCH = "v"
ROTOR4_NOTCH = "j"
ROTOR5_NOTCH = "z"
ROTOR6_NOTCH = "m"			# 6-8はmまたはzらしい
ROTOR7_NOTCH = "m"
ROTOR8_NOTCH = "m"

# プラグボードとリフレクタは回転しないローターとみなしてよい
PLUGBOARD_WIRE = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"	# 全ての配線を入れ換えることは
						# ないらしい。付属プラグボード
						# ワイヤが12本らしい
						# 全部入れ換えると組合せの
						# 場合の数が減る
REFLECTOR1_WIRE = "EJMZALYXVBWFCRQUONTSPIKHGD"
REFLECTOR2_WIRE = "YRUHQSLDPXNGOKMIEBFZCWVJAT"
REFLECTOR3_WIRE = "FVPJIAOYEDRZXWGCTKUQSBNMHL"


# ローター。入力xを出力Xに1:1変換する。変換テーブルはコメントを参照
# プラグボードとリフレクタは回転しないローターとみなす。
class Rotor
    def initialize(connection, notch = nil, ring = nil)
        @connection = connection.split(//)	# 配線というか変換表
        @ring = nil
        if ring != nil			# リング設定
            @ring = ring
            i = 0
            while i < ring
                @connection.push(@connection.shift)
                i += 1
            end
        end
            
        @notch = nil
        if notch != nil
            @notch = @connection[notch.upcase.ord - 'A'.ord]
            # ローターを次に回す位置
            # notch == q であればローターqからrに動くときに次の
            # ローターを回す。
        end

#        p @connection
#        p "notch = #{@notch}"
    end
    
    # ローターを1段階回す。プラグボードとリフレクタでは使わない
    # 返り値	true	1周回った
    #		false	その他
    def step
        turnnext = false
        if @connection[0] == @notch
            turnnext = true
        end
        @connection.push(@connection.shift)
        
        return turnnext
    end

    # ローターの初期位置を設定
    def set_init_pos(pos)
        i = pos.upcase.ord - 'A'.ord
        while i > 0
            step()
            i -= 1
        end
    end
    
    # 順方向(ローター1→ローター2、ローター2→ローター3、...)の変換
    def forward_enc(p)
        p = p.upcase
        return @connection[p.ord - 'A'.ord]
    end

    # 逆方向(ローター3→ローター2、ローター2→ローター1、...)の変換
    # リフレクタでは使わない
    def reverse_enc(p)
        p = p.upcase
        pos = 0
        @connection.each do |c|
            if p == c
                return (pos + 'A'.ord).chr
            end
            pos += 1
        end
    end
end


# エニグマエミュレータ
class Enigma
    def initialize(plugboard, rotor_wire, rotor_notch, rotor_ring, reflector)
        @plugboard = Rotor.new(plugboard)
        @rotor = Array.new
        if rotor_wire.length != rotor_notch.length or
          rotor_notch.length != rotor_ring.length
            print "sizes of rotor wire, notch and ring differ!\n"
            exit
        end
        i = 0
        while i < rotor_wire.length
            @rotor.push(Rotor.new(rotor_wire[i], rotor_notch[i], rotor_ring[i]))
            i += 1
        end
        @reflector = Rotor.new(reflector)
    end

    attr_reader :rotor


    # 暗号化と復号化の計算本体
    def convert(input_str)
        output_str = ""
        input_str.each_char do |c|
            c.upcase!
            if c.match(/[A-Z]/)
                a1 = @plugboard.forward_enc(c)
                a2 = @rotor[0].forward_enc(a1)
                a3 = @rotor[1].forward_enc(a2)
                a4 = @rotor[2].forward_enc(a3)
                a5 = @reflector.forward_enc(a4)
                a6 = @rotor[2].reverse_enc(a5)
                a7 = @rotor[1].reverse_enc(a6)
                a8 = @rotor[0].reverse_enc(a7)
                a9 = @plugboard.reverse_enc(a8)
                output_str = output_str + a9
                if @rotor[0].step
                    if @rotor[1].step
                        @rotor[2].step
                    end
                end
            else
                output_str = output_str + c
            end
        end

        return output_str
    end

    # 暗号化。大文字で返すだけ
    def encrypt(plaintext)
        return convert(plaintext).upcase
    end

    # 復号化。小文字で返すだけで計算自体は暗号化と共通
    def decrypt(ciphertext)
        return convert(ciphertext).downcase
    end
    

end


